#!/usr/bin/env python

import argparse
import csv
import sys
from collections import Counter, defaultdict
from itertools import islice
from tabulate import tabulate
from tqdm import tqdm

from .ud import download_hdt
from .. import Lemmatizer
from ..log import configure_logging


ud_to_dwdsmor_pos = {
    "ADJA": {"ORD", "FRAC", "ADJ", "CARD", "INDEF"},
    "ADJD": {"ADJ"},
    "ADV": {"ADJ", "ADV", "PROADV"},
    "APPO": {"POSTP"},
    "APPR": {"PREP"},
    "APPRART": {"PREPART"},
    "APZR": {"ADV", "POSTP", "PREP"},
    "ART": {"ART"},
    "CARD": {"FRAC", "CARD"},
    "FM": {"NN", "ADJ", "PREP"},
    "ITJ": {"INTJ", "ADV"},
    "KOKOM": {"CONJ"},
    "KON": {"CONJ"},
    "KOUI": {"CONJ"},
    "KOUS": {"CONJ"},
    "NE": {"NPROP"},
    "NN": {"NN", "ADJ", "NPROP"},
    "PDAT": {"DEM"},
    "PDS": {"DEM"},
    "PIAT": {"ADJ", "ART", "INDEF", "DEM"},
    "PIDAT": {"ADJ", "INDEF", "DEM"},
    "PIS": {"ADJ", "INDEF", "DEM"},
    "PPER": {"PPRO"},
    "PPOSAT": {"POSS"},
    "PPOSS": {"POSS"},
    "PRELAT": {"REL"},
    "PRELS": {"REL"},
    "PRF": {"PPRO"},
    "PROAV": {"PROADV"},
    "PROP": {"ADV", "PROADV"},
    "PTKA": {"PTCL"},
    "PTKANT": {"INTJ", "PTCL"},
    "PTKNEG": {"PTCL"},
    "PTKVZ": {"ADV", "ADJ", "PREP", "PREV", "V"},
    "PTKZU": {"PTCL"},
    "PWAT": {"WPRO"},
    "PWAV": {"ADV", "PROADV"},
    "PWS": {"WPRO"},
    "TRUNC": {"NN", "V"},
    "VAFIN": {"V"},
    "VAIMP": {"V"},
    "VAINF": {"V"},
    "VAPP": {"V"},
    "VMFIN": {"V"},
    "VMINF": {"V"},
    "VMPP": {"V"},
    "VVFIN": {"V"},
    "VVIMP": {"V"},
    "VVINF": {"V"},
    "VVIZU": {"V"},
    "VVPP": {"V"},
    "XY": {"XY", "NN"},
    "$.": {"PUNCT"},
    "$,": {"PUNCT"},
    "$(": {"PUNCT"},
}

coverage_headers = [
    "POS",
    "Types",
    "Tokens",
    "% Types",
    "% Tokens",
    "% Lemmatized Types",
    "% Lemmatized Tokens",
]


def compute_coverage(round_digits=4, limit=None, debug=False, quiet=False):
    lemmatizer = Lemmatizer()

    def hdt_tokens():
        for sentence in download_hdt():
            for token in sentence:
                lemma = token.get("lemma") or "NULL"
                if lemma in {"unknown", "NULL"}:
                    continue  # remove unknown lemmata
                form = token.get("form") or "-"
                if "-" in form:
                    continue  # remove Bindestrich-Komposita
                pos = token.get("xpos") or "XY"
                yield form, lemma, pos

    tokens = hdt_tokens()
    if limit is not None:
        tokens = islice(tokens, max(0, limit))
    if debug:
        report = csv.writer(sys.stdout)
        report.writerow(("Form", "Lemma", "POS", "Lemmatized"))
    if not debug or quiet:
        tokens = tqdm(tokens, unit="token", desc="Lemmatizing UD/de-hdt", total=limit)
    matches = defaultdict(Counter)
    mismatches = defaultdict(Counter)
    for token in tokens:
        form, lemma, xpos = token
        pos_candidates = ud_to_dwdsmor_pos.get(xpos, set())
        is_match = lemmatizer(form, pos=pos_candidates) is not None
        if not is_match and lemmatizer(lemma) is not None:
            # skip tokens where we can analyze the given lemma but not the form:
            # compounds are lemmatized to their basic words in German-UD/HDT
            continue
        if debug:
            report.writerow((form, lemma, xpos, 1 if is_match else 0))
        registry = matches if is_match else mismatches
        registry[xpos][form] += 1

    coverage = []

    total_tokens = 0
    total_types = 0
    total_type_matches = 0
    total_token_matches = 0
    for registry in (matches, mismatches):
        for _pos_tag, types in registry.items():
            for _type, token_count in types.items():
                total_types += 1
                total_tokens += token_count
    pos_tags = sorted(set(matches.keys()).union(mismatches.keys()))
    for pos_tag in pos_tags:
        tag_tokens = 0
        tag_types = 0
        tag_token_matches = 0
        tag_type_matches = 0
        for registry, is_match in ((matches, True), (mismatches, False)):
            for _type, token_count in registry.get(pos_tag, {}).items():
                tag_types += 1
                tag_tokens += token_count
                if is_match:
                    total_type_matches += 1
                    total_token_matches += token_count
                    tag_type_matches += 1
                    tag_token_matches += token_count
        coverage.append(
            [
                pos_tag,
                tag_types,
                tag_tokens,
                round(tag_types / total_types, round_digits),
                round(tag_tokens / total_tokens, round_digits),
                round(tag_type_matches / tag_types, round_digits),
                round(tag_token_matches / tag_tokens, round_digits),
            ]
        )
    coverage.insert(
        0,
        [
            "",
            total_types,
            total_tokens,
            1,
            1,
            round(total_type_matches / total_types, round_digits),
            round(total_token_matches / total_tokens, round_digits),
        ],
    )
    return coverage


if __name__ == "__main__":
    arg_parser = argparse.ArgumentParser(
        prog=__package__,
        description="Benchmarks coverage of DWDSmor against German-UD/HDT.",
    )
    arg_parser.add_argument(
        "-d", "--debug", help="print coverage data to stdout", action="store_true"
    )
    arg_parser.add_argument(
        "-q", "--quiet", help="do not report progress", action="store_true"
    )
    arg_parser.add_argument(
        "--limit",
        help="number of tokens used for the benchmark (default: all)",
        type=int,
    )
    args = arg_parser.parse_args()
    configure_logging()

    colalign = ["left", "right", "right", "right", "right", "right", "right"]

    def pct(v):
        return "{:.2%}".format(v)

    def count(v):
        return "{:,d}".format(v)

    def format_coverage(pos_coverage, is_total=False):
        (
            pos,
            types_total,
            tokens_total,
            types_ratio,
            token_ratio,
            type_coverage,
            token_coverage,
        ) = pos_coverage
        return [
            "Σ" if is_total else pos,
            count(types_total),
            count(tokens_total),
            pct(types_ratio),
            pct(token_ratio),
            pct(type_coverage),
            pct(token_coverage),
        ]

    coverage = compute_coverage(
        limit=args.limit,
        debug=args.debug,
        quiet=args.quiet,
    )

    if not args.debug:
        print(
            tabulate(
                [
                    format_coverage(coverage, n == 0)
                    for n, coverage in enumerate(coverage)
                ],
                tablefmt="tsv",
                headers=coverage_headers,
                colalign=colalign,
            )
        )



import csv
from collections import defaultdict

import dwdsmor
from dwdsmor.build.traversal import traverse
from dwdsmor.traversal import Traversal

from .conftest import if_dwds_available, test_dir


@if_dwds_available
def test_analysis(snapshot):
    analyzer = dwdsmor.analyzer()
    data_dir = test_dir / "gesetze"
    data_files = sorted(data_dir.rglob("*.tok"))
    tokens = [t for f in data_files for t in f.read_text().splitlines()]
    result = [(t, a.spec) for t in tokens for a in analyzer.analyze(t)]
    assert snapshot == result


@if_dwds_available
def test_generation(snapshot):
    test_cases_file = test_dir / "inflection_test_cases.csv"

    traversals = defaultdict(list)
    for surface, spec in traverse(dwdsmor.default_automata_dir / "index.a"):
        t = Traversal.parse(spec)
        traversals[(t.analysis, t.pos)].append((t, surface))

    result = []

    with test_cases_file.open("rt") as f:
        for n, (pos, inflection_classes, lemma) in enumerate(csv.reader(f)):
            if n == 0:
                continue
            lemma_traversals = sorted(
                (t.spec, surface) for t, surface in traversals.get((lemma, pos), [])
            )
            result.append((pos, inflection_classes, lemma, *lemma_traversals))
    assert result == snapshot



import dwdsmor


def test_automata():
    for automaton_type in dwdsmor.automaton_types:
        analyzer = dwdsmor.analyzer(automaton_type=automaton_type)
        generator = dwdsmor.generator(automaton_type=automaton_type)
        for lemma, inflected in (("testen", "testet"),):
            for analyzed in analyzer.analyze(inflected):
                assert analyzed.analysis == lemma
                for generated in generator.generate(analyzed.spec):
                    assert inflected == generated.surface
                    break



def is_current(target, sources):
    "Check whether a target file is current, given the mtime of its sources."
    if not target.exists():
        return False
    target_m_time = target.stat().st_mtime
    return not any((target_m_time < source.stat().st_mtime for source in sources))



from spacy.language import Language
from spacy.tokens.token import Token

import dwdsmor.tag.hdt as hdt

from . import lemmatizer

Token.set_extension("dwdsmor", default=None)


def morph(token_morph, k):
    v = ",".join(token_morph.get(k))
    return v if v else None


class Component:
    def __init__(self):
        self.lemmatizer = lemmatizer()

    def __call__(self, doc):
        for token in doc:
            token_morph = token.morph
            token_criteria = hdt.criteria(
                token.tag_,
                morph(token_morph, "Number"),
                morph(token_morph, "Gender"),
                morph(token_morph, "Case"),
                morph(token_morph, "Person") or "UnmPers"
                if token.dep_ == "compound:prt"
                else None,
                morph(token_morph, "Tense"),
                morph(token_morph, "Degree"),
                morph(token_morph, "Mood"),
                morph(token_morph, "VerbForm"),
                "SEP" if token.dep_ == "compound:prt" else None,
            )
            token._.dwdsmor = self.lemmatizer(token.text, **token_criteria)
        return doc


@Language.factory("dwdsmor")
def create_component(nlp: Language, name: str):
    return Component()



from itertools import islice

import spacy
from pytest import fixture

import dwdsmor.spacy  # noqa
from dwdsmor.build.ud import download_gsd

from .conftest import if_dwds_available


@fixture(scope="module")
def nlp():
    nlp = spacy.load("de_zdl_lg")
    nlp.add_pipe("dwdsmor")
    return nlp


@fixture(scope="module")
def sentences():
    sentences = (s.metadata["text"] for s in download_gsd())
    return tuple(islice(sentences, 10))


@if_dwds_available
def test_lemmatisation(nlp, sentences, snapshot):
    docs = nlp.pipe(sentences)
    tokens = (
        (t.text, t.tag_, t.lemma_, t._.dwdsmor.analysis)
        for d in docs
        for t in d
        if t._.dwdsmor and t.lemma_ != t._.dwdsmor.analysis
    )
    assert tuple(tokens) == snapshot


@if_dwds_available
def test_particle_lemmatization(nlp):
    sentences = [
        "Sie nimmt nicht an der Wahl teil.",
        "Wir arbeiten beständig weiter.",
    ]
    docs = nlp.pipe(sentences)
    tokens = [[t._.dwdsmor.analysis for t in doc] for doc in docs]
    expected = [
        ["sie", "nehmen", "nicht", "an", "die", "Wahl", "teil", "."],
        ["wir", "arbeiten", "beständig", "weiter", "."],
    ]
    assert tokens == expected



import gzip
from io import BytesIO
from pathlib import Path
from shutil import copyfileobj

import dwdswb

project_dir = Path(__file__).parent.parent.parent
lexicon_dir = project_dir / "lexicon" / "dwds" / "wb"
lexicon_file = lexicon_dir / "dwdswb.xml"


def main():
    lexicon_dir.mkdir(parents=True, exist_ok=True)
    with gzip.open(BytesIO(dwdswb.dataset())) as src, lexicon_file.open("wb") as dest:
        copyfileobj(src, dest)


if __name__ == "__main__":
    main()



import logging
import os
import sys


def configure_logging():
    logging.basicConfig(
        format="%(asctime)s - %(name)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        stream=sys.stderr,
        level=(logging.DEBUG if os.getenv("DEBUG") else logging.INFO),
    )



from collections import OrderedDict
from functools import cache

pos_map = {
    "$(": {"PUNCT"},
    "$,": {"PUNCT"},
    "$.": {"PUNCT"},
    "ADJA": {"ADJ", "CARD", "INDEF", "ORD"},
    "ADJD": {"ADJ"},
    "ADV": {"ADV"},
    "APPO": {"POSTP"},
    "APPR": {"PREP"},
    "APPR_ART": {"PREPART"},
    "APZR": {"POSTP", "PREP"},
    "ART": {"ART"},
    "CARD": {"CARD"},
    "FM": {"FM"},  # ?
    "ITJ": {"INTJ"},
    "KOKOM": {"CONJ"},
    "KON": {"CONJ"},
    "KOUI": {"CONJ"},
    "KOUS": {"CONJ"},
    "NE": {"NN", "NPROP"},
    "NN": {"NN", "NPROP"},
    "PDAT": {"DEM"},
    "PDS": {"DEM"},
    "PIAT": {"INDEF"},
    "PIDAT": {"INDEF"},
    "PIS": {"INDEF"},
    "PPER": {"PPRO"},
    "PPOSAT": {"POSS"},
    "PPOSS": {"POSS"},
    "PRELAT": {"REL"},
    "PRELS": {"REL"},
    "PRF": {"PPRO"},
    "PROAV": {"ADV", "PROADV"},
    "PTKA": {"PTCL"},
    "PTKANT": {"INTJ", "PTCL"},
    "PTKNEG": {"PTCL"},
    "PTKVZ": {"ADV", "ADJ", "PREP", "PREV", "V"},
    "PTKZU": {"PTCL"},
    "PWAT": {"WPRO"},
    "PWAV": {"ADV"},
    "PWS": {"WPRO"},
    "TRUNC": {"TRUNC"},  # ?
    "VAFIN": {"V"},
    "VAIMP": {"V"},
    "VAINF": {"V"},
    "VAPP": {"V"},
    "VMFIN": {"V"},
    "VMINF": {"V"},
    "VMPP": {"V"},
    "VVFIN": {"V"},
    "VVIMP": {"V"},
    "VVINF": {"V"},
    "VVIZU": {"V"},
    "VVPP": {"V"},
    "XY": {"XY"},  # ?
}

number_map = {
    "Sing": {"Sg"},
    "Plur": {"Pl"},
}


gender_map = {
    "Masc,Neut": {"Masc", "Neut"},
    "Neut": {"Neut"},
    "Fem": {"Fem"},
    "Masc": {"Masc"},
}

case_map = {
    "Nom": {"Nom"},
    "Gen": {"Gen"},
    "Dat": {"Dat"},
    "Acc": {"Acc"},
}

person_map = {
    "1": {"1"},
    "2": {"2"},
    "3": {"3"},
}

tense_map = {
    "Past": {"Past"},
    "Pres": {"Pres"},
}


degree_map = {
    "Cmp": {"Comp"},
    "Sup": {"Sup"},
    "Pos": {"Pos"},
}

mood_map = {
    "Ind": {"Ind"},
    "Imp": {"Imp"},
}

# VerbForm
nonfinite_map = {
    "Part": {"Part"},
    "Inf": {"Inf"},
}

syninfo_map = {"SEP": {"SEP"}}


def criterion(k, v, mapping):
    return (k, mapping.get(v, {v}) if v else None)


@cache
def criteria(
    pos, number, gender, case, person, tense, degree, mood, nonfinite, syninfo
):
    return OrderedDict(
        (
            criterion("pos", pos, pos_map),
            criterion("number", number, number_map),
            criterion("gender", gender, gender_map),
            criterion("case", case, case_map),
            criterion("person", person, person_map),
            criterion("tense", tense, tense_map),
            criterion("degree", degree, degree_map),
            criterion("mood", mood, mood_map),
            criterion("nonfinite", nonfinite, nonfinite_map),
            criterion("syninfo", syninfo, syninfo_map),
        )
    )



from dwdsmor.tag import prev_boundary
from dwdsmor.traversal import Traversal


def inflected(spec, generated):
    traversal = Traversal.parse(
        spec, visible_boundaries=prev_boundary, boundary_tag="|"
    )
    if traversal.pos == "V" and traversal.nonfinite is None and prev_boundary in spec:
        parts = traversal.analysis.split("|")
        parts = parts[:-1]
        parts_length = sum(len(p) for p in parts)
        base = generated[parts_length:]
        return f"{base} {' '.join(parts)}"
    return generated
